body {
    font-family: 'Nunito', sans-serif, 'Helvetica', sans-serif;
    text-align: center;
    margin-bottom: 4%;
}

h1 {
    font-size: 35px;
    margin-top: 4%;
    margin-bottom: 0;
    font-weight: normal;
}

h2 {
    margin: 0;
}

h3 {
    margin: 10px;
    font-weight: normal;
    font-size: 20px;
}

button {
    background-color: transparent;
    cursor: pointer;
    border-style: none;
}

button:hover{
    color: lightgrey;
}

#completionChart {
   height: 150px;
}

#listOfEvaluationsDiv{
    margin-top: 30px;
    display: inline-block;
}

.evaluationDiv{
    text-align: left;
    overflow: hidden;
    margin-top: 20px;
}

.detailedResultsDiv{
    display: none;
    text-align: left;
    overflow: hidden;
    margin-top: 20px;
    margin-left: 30px;
}

.innerResultDiv{
    margin-top: 5px;
}

.qualityStandardTextDiv{
    margin-left: 35px;
    display: inline-block;
    vertical-align: middle;
    font-size: 20px;
}

.innerQualityStandardTextDiv{
    margin-left: 35px;
    display: inline-block;
    vertical-align: middle;
    font-size: 18px;
}

.statusIconDiv{
    display: inline-block;
    vertical-align: middle;
}

.statusIcon {
    width: 40px;
    vertical-align: middle;
}

.innerStatusIcon {
    width: 24px;
    vertical-align: middle;
}

.errorMsg {
    display: none;
    margin-left: 80px;
    font-size: 16px;
    color: red;
}