NAME_KEY_IDENTIFIER = "name"
METRIC_KEY_IDENTIFIER = "metric"
OUTPUT_NAME_KEY_IDENTIFIER = "output_name"
THRESHOLD_KEY_IDENTIFIER = "threshold"
CONDITION_KEY_IDENTIFIER = "condition"
DIMENSION_KEY_IDENTIFIER = "dimension"
DIMENSION_VALUE_KEY_IDENTIFIER = "dimension_value"

DETAILED_RESULTS_KEY_IDENTIFIER = "detailed_results"
EVALUATIONS_RESULTS_KEY_IDENTIFIER = "evaluations_results"
MESSAGE_KEY_IDENTIFIER = "message"
METRIC_VALUE_KEY_IDENTIFIER = "metric_value"
STATUS_KEY_IDENTIFIER = "status"
PASS_COUNT_KEY_IDENTIFIER = "pass_count"
FAIL_COUNT_KEY_IDENTIFIER = "fail_count"
ERROR_COUNT_KEY_IDENTIFIER = "error_count"
